@@ -1,5 +1,6 @@
 # -*- coding: utf-8 -*-
 
+import datetime
 import os
 import sys
 
@@ -11,8 +12,8 @@ sys.path.insert(0, os.path.abspath('../..'))
 extensions = [
     'sphinx.ext.autodoc',
     'sphinx.ext.doctest',
+    'sphinx.ext.extlinks',
     'sphinx.ext.inheritance_diagram',
-    'sphinx.ext.intersphinx',
     'sphinx.ext.viewcode',
     'oslosphinx'
 ]
@@ -36,7 +37,8 @@ exclude_patterns = ['_build']
 
 # General information about the project.
 project = u'TaskFlow'
-copyright = u'2013-2014, OpenStack Foundation'
+copyright = u'%s, OpenStack Foundation' % datetime.date.today().year
+source_tree = 'http://git.openstack.org/cgit/openstack/taskflow/tree'
 
 # If true, '()' will be appended to :func: etc. cross-reference text.
 add_function_parentheses = True
@@ -51,6 +53,11 @@ pygments_style = 'sphinx'
 # Prefixes that are ignored for sorting the Python module index
 modindex_common_prefix = ['taskflow.']
 
+# Shortened external links.
+extlinks = {
+    'example': (source_tree + '/taskflow/examples/%s.py', ''),
+    'pybug': ('http://bugs.python.org/issue%s', ''),
+}
 
 # -- Options for HTML output --------------------------------------------------
 
@@ -76,9 +83,6 @@ latex_documents = [
      'OpenStack Foundation', 'manual'),
 ]
 
-# Example configuration for intersphinx: refer to the Python standard library.
-intersphinx_mapping = {'http://docs.python.org/': None}
-
 # -- Options for autoddoc ----------------------------------------------------
 
 # Keep source order
