@@ -16,9 +16,8 @@
 
 import testtools
 
-from concurrent import futures
-
 from taskflow import test
+from taskflow.types import futures
 from taskflow.utils import async_utils as au
 from taskflow.utils import eventlet_utils as eu
 
@@ -30,7 +29,7 @@ class WaitForAnyTestsMixin(object):
         def foo():
             pass
 
-        with self.executor_cls(2) as e:
+        with self._make_executor(2) as e:
             fs = [e.submit(foo), e.submit(foo)]
             # this test assumes that our foo will end within 10 seconds
             done, not_done = au.wait_for_any(fs, 10)
@@ -57,31 +56,14 @@ class WaitForAnyTestsMixin(object):
 @testtools.skipIf(not eu.EVENTLET_AVAILABLE, 'eventlet is not available')
 class AsyncUtilsEventletTest(test.TestCase,
                              WaitForAnyTestsMixin):
-    executor_cls = eu.GreenExecutor
-    is_green = True
-
-    def test_add_result(self):
-        waiter = eu._GreenWaiter()
-        self.assertFalse(waiter.event.is_set())
-        waiter.add_result(futures.Future())
-        self.assertTrue(waiter.event.is_set())
-
-    def test_add_exception(self):
-        waiter = eu._GreenWaiter()
-        self.assertFalse(waiter.event.is_set())
-        waiter.add_exception(futures.Future())
-        self.assertTrue(waiter.event.is_set())
-
-    def test_add_cancelled(self):
-        waiter = eu._GreenWaiter()
-        self.assertFalse(waiter.event.is_set())
-        waiter.add_cancelled(futures.Future())
-        self.assertTrue(waiter.event.is_set())
+    def _make_executor(self, max_workers):
+        return futures.GreenThreadPoolExecutor(max_workers=max_workers)
 
 
 class AsyncUtilsThreadedTest(test.TestCase,
                              WaitForAnyTestsMixin):
-    executor_cls = futures.ThreadPoolExecutor
+    def _make_executor(self, max_workers):
+        return futures.ThreadPoolExecutor(max_workers=max_workers)
 
 
 class MakeCompletedFutureTest(test.TestCase):
@@ -91,3 +73,16 @@ class MakeCompletedFutureTest(test.TestCase):
         future = au.make_completed_future(result)
         self.assertTrue(future.done())
         self.assertIs(future.result(), result)
+
+    def test_make_completed_future_exception(self):
+        result = IOError("broken")
+        future = au.make_completed_future(result, exception=True)
+        self.assertTrue(future.done())
+        self.assertRaises(IOError, future.result)
+        self.assertIsNotNone(future.exception())
+
+
+class AsyncUtilsSynchronousTest(test.TestCase,
+                                WaitForAnyTestsMixin):
+    def _make_executor(self, max_workers):
+        return futures.SynchronousExecutor()
